/*
 * The I/O retry and loop functions for the hasher-priv project.
 *
 * Copyright (C) 2003-2022  Dmitry V. Levin <ldv@altlinux.org>
 * All rights reserved.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* Code in this file may be executed with caller or child privileges. */

#include "io_loop.h"
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>

/* This function may be executed with caller privileges. */
ssize_t
read_retry(int fd, void *buf, size_t count)
{
	return TEMP_FAILURE_RETRY(read(fd, buf, count));
}

/* This function may be executed with caller privileges. */
ssize_t
write_retry(int fd, const void *buf, size_t count)
{
	return TEMP_FAILURE_RETRY(write(fd, buf, count));
}

/* This function may be executed with child privileges. */
ssize_t sendmsg_retry(int fd, const struct msghdr *msg, int flags)
{
	return TEMP_FAILURE_RETRY(sendmsg(fd, msg, flags));
}

/* This function may be executed with caller privileges. */
ssize_t recvmsg_retry(int fd, struct msghdr *msg, int flags)
{
	return TEMP_FAILURE_RETRY(recvmsg(fd, msg, flags));
}

/* This function may be executed with child privileges. */
ssize_t
read_loop(int fd, char *buffer, size_t count)
{
	ssize_t offset = 0;

	while (count > 0) {
		ssize_t block = read_retry(fd, &buffer[offset], count);

		if (block <= 0)
			return offset ? : block;
		offset += block;
		count -= (size_t) block;
	}
	return offset;
}

/* This function may be executed with caller privileges. */
ssize_t
write_loop(int fd, const char *buffer, size_t count)
{
	ssize_t offset = 0;

	while (count > 0) {
		ssize_t block = write_retry(fd, &buffer[offset], count);

		if (block <= 0)
			return offset ? : block;
		offset += block;
		count -= (size_t) block;
	}
	return offset;
}
